require 'yaml'
require 'field'
require 'items/explosion'
require 'items/wall'
require 'items/blowable_wall'
require 'items/beamer'

class Level
  attr_accessor :name, :map, :start_points, :file, :editing
  
  def initialize(name, map, start_points = [], beamer = nil, editing = false)
    @name = name
    @editing = editing
    @file = nil
    
    i_row = -1
    @map = map.map do |row|
      i_row += 1
      i_field = -1
      row.map do |item|
        i_field += 1
        f = Field.new V(i_field * 40, i_row * 40),
                      Rubygame::Rect.new(i_field*40, i_row*40, 40, 40),
                      V(i_field, i_row)
        item.to_a.each do |char|
          case char
            when 'w'
              Wall.new(f)
            when 'b'
              BlowableWall.new(f)
          end
        end
        f
      end
    end
    
    if beamer
      beamer.each do |b|
        next unless b.size == 2 && b[0].size == 2 && b[1].size == 2
        b1 = Beamer.new(self[b[0]])
        b2 = Beamer.new(self[b[1]])
        b1.destination = V(b[1])*40
        b2.destination = V(b[0])*40
        self[b[0]].items.push b1
        self[b[1]].items.push b2
      end
    end
    
    @start_points = start_points
  end
  
  def self.load(file)
    yml = YAML.load(File.open(file).read)
    @@currently_loading = yml.merge(Settings.gameplay.get(:level_defaults))
    
    distr = []
    @@currently_loading["items"]["distribution"].each do |item, n|
      distr += [Kernel.const_get("#{item}Item")] * n
    end
    @@currently_loading["items"]["distribution"] = distr
    
    l = Level.new yml["name"],
                  yml["map"],
                  yml["start_points"].map{ |sp| V(sp) },
                  yml["beamer"]
    
    @@currently_loading = nil
    l.file = file
    return l
  end
  
  def save(file = nil)
    @file = file unless file.nil?
    File.open(@file, 'w') do |file|
      YAML.dump({
        "name"         => @name,
        "map"          => @map.map do |row|
                            row.map do |field|
                              field.get_item_chars
                            end
                          end,
        "start_points" => @start_points.map(&:to_a)
      }, file)
      file.close
    end
  end
  
  def self.load_all
    (Dir.entries("levels") - [".", ".."]).map do |file|
      Level.load "levels/#{file}"
    end
  end
  
  def self.count_levels
    (Dir.entries("levels") - [".", ".."]).size
  end
  
  
  def draw(srf, position)
    srf.fill [200, 200, 200], Rubygame::Rect.new(position.to_a, [@map[0].size*40, @map.size*40])
    @map.each do |row|
      row.each do |field|
        unless field.nil?
          field.draw srf, position
        end
      end
    end
    
    if @editing
      # draw start points
      spsrf = Rubygame::Surface.new [40, 40]
      spsrf.fill [255, 0, 0]
      spsrf.alpha = 150
      start_points.each_with_index do |sp, i|
        spsrf.blit srf, (sp*40).to_a
        t = Game.font.render("#{i+1}", false, [255, 255, 255])
        t.blit srf, (sp*40 + V(20-t.w/2, 20-t.h/2)).to_a
      end
    end
  end
  
  def update(dtime)
    @explosions_created = []
    
    @map.each { |row| row.each { |field| field.update(dtime) } }
    
    if !@editing &&
       !Settings.gameplay.sudden_death_after.nil? &&
       Settings.gameplay.sudden_death_after <= Game.current_controller.game_timer
      @sudden_death_countdown ||= rand(500)
      @sudden_death_countdown -= dtime
      
      if @sudden_death_countdown <= 0
        @sudden_death_countdown = nil
        w = @map[0].size
        h = @map.size
        field = self[V(rand(w), rand(h))]
        until field.blowable?
          field = self[V(rand(w), rand(h))]
        end
        Explosion.new(field)
      end
    end
  end
  
  def columns
    @map[0].size
  end
  def rows
    @map.size
  end
  
  def width
    columns * 40
  end
  def height
    rows * 40
  end
  
  def [](pos)
    @map[pos[1]][pos[0]]
  end
  
  def field_at(pos, center = true)
    pos += V(40/2, 40/2) if center
    V((pos.x / 40).floor, (pos.y / 40).floor)
  end
  
  def entered_field(pos1, pos2, v)
    # check if player entered another field
    if v.x < 0 && (pos2.x / 40).floor != (pos1.x / 40).floor
      f = V((pos2.x / 40).floor, (pos2.y / 40).floor)
    elsif  v.x > 0 && ((pos2.x+40-1) / 40).floor != ((pos1.x+40-1) / 40).floor
      f = V(((pos2.x+40-1) / 40).floor, (pos2.y / 40).floor)
    elsif  v.y < 0 && (pos2.y / 40).floor != (pos1.y / 40).floor
      f = V((pos2.x / 40).floor, (pos2.y / 40).floor)
    elsif  v.y > 0 && ((pos2.y+40-1) / 40).floor != ((pos1.y+40-1) / 40).floor
      f = V((pos2.x / 40).floor, ((pos2.y+40-1) / 40).floor)
    end
    
    unless f.nil?
      @map[f.y][f.x]
    else
      nil
    end
  end
  
  def find_collisions(pos, margin = 1)
    points = [
      V(pos) + [margin, margin],
      V(pos) + [40 - margin, margin],
      V(pos) + [margin, 40 - margin],
      V(pos) + [40 - margin, 40 - margin]
    ]
    
    points.map do |p|
      self[self.field_at(p, false)]
    end.uniq
  end
  
  def create_explosion(field, player)
    explosion_id = Explosion.new(field, player).explosion_id
    
    [V(1, 0), V(0, 1), V(-1, 0), V(0, -1)].each do |dir|
      blocked = false
      player.explosive_force.times do |n|
        next if blocked
        coords = field.coords + dir*(n+1)
        if coords.x >= 0 && coords.x < @map[0].size && coords.y >= 0 && coords.y < @map.size
          f = self[field.coords + dir*(n+1)]
          blocked = f.blocks_explosion?
          if f.blowable?
            already_exploding = @explosions_created.include?(coords)
            Explosion.new f, player, explosion_id, !already_exploding
            @explosions_created.push coords unless already_exploding
          end
        end
      end
    end
  end
  
  # checks if the given coordinates lie inside the level
  def in_level?(coords)
    coords.x >= 0 && coords.x < @map[0].size && coords.y >= 0 && coords.y < @map.size
  end
  
  
  @@currently_loading = nil
  def self.currently_loading
    @@currently_loading
  end
end
